theory Session4 imports Main begin

section {* Sets *}
text {* A set is defined by its characteristic function. Have a look
at the type of sets. *}
typ "'a set"

text {* An element `x' is a member of the set `S', if the characteristic
function of `S' returns `True'. *}
thm mem_def

subsection {* Basic Operations *}
subsubsection {* Intersection *}
text {* Notation *}
term "A \<inter> B"
text {* ASCII *}
term "A Int B"
text {* Natural deduction rules. *}
thm IntI IntD1 IntD2 IntE

subsubsection {* Union *}
text {* Notation *}
term "A \<union> B"
text {* ASCII *}
term "A Un B"
text {* Natural deduction rules. *}
thm UnI1 UnI2 UnE

subsection {* Set Notation *}
text {* The empty set. *}
term "{}"
text {* The universal set. *}
term "UNIV"
text {* Singleton sets. *}
term "{x}"
text {* Insertion. *}
term "insert x A"
thm insert_is_Un

text {* An example proof. *}
lemma "A \<inter> (B \<union> C) = (A \<inter> B) \<union> (A \<inter> C)" (is "?S = ?T")
proof (rule equalityI)
  show "?S \<subseteq> ?T"
  proof (rule subsetI)
    fix x assume "x \<in> ?S"
    hence A: "x \<in> A" by (rule IntD1)
    from `x \<in> ?S` have "x \<in> B \<union> C" by (rule IntD2)
    thus "x \<in> ?T"
    proof (rule UnE)
      assume "x \<in> B" with A have "x \<in> A \<inter> B" by (rule IntI)
      thus ?thesis by (rule UnI1)
    next
      assume "x \<in> C" with A have "x \<in> A \<inter> C" by (rule IntI)
      thus ?thesis by (rule UnI2)
    qed
  qed
next
  show "?T \<subseteq> ?S"
  proof (rule subsetI)
    fix x assume "x \<in> ?T"
    thus "x \<in> ?S"
    proof (rule UnE)
      assume "x \<in> A \<inter> B"
      hence "x \<in> A" by (rule IntD1)
      from `x \<in> A \<inter> B` have "x \<in> B" by (rule IntD2)
      hence "x \<in> B \<union> C" by (rule UnI1)
      with `x \<in> A` show ?thesis by (rule IntI)
    next
      assume "x \<in> A \<inter> C"
      hence "x \<in> A" by (rule IntD1)
      from `x \<in> A \<inter> C` have "x \<in> C" by (rule IntD2)
      hence "x \<in> B \<union> C" by (rule UnI2)
      with `x \<in> A` show ?thesis by (rule IntI)
    qed
  qed
qed

text {* And now shorter. *}
lemma "A \<inter> (B \<union> C) = (A \<inter> B) \<union> (A \<inter> C)" by blast

text {* Examples of set comprehension. *}
text {* The set of all elements satisfying P. *}
term "{x. P x}"
text {* The set of all pairs where the first component comes
from A and the second component form B. *}
term "{(x,y) | x y. x \<in> A \<and> y \<in> B}"

subsection{* Binding Operators For Sets *}
term "\<forall>x\<in>A. P x"
term "\<exists>x\<in>A. P x"
thm ballI bspec bexI bexE

subsection {* Relations *}
text {* A relation is represented by a set of pairs. *}
text {* The identity relation. *}
term "Id"
thm Id_def

text {* Composition of relations. *}
term "r O s"
thm rel_comp_def

text {* Converse of a relation. *}
term "r^-1"
thm converse_iff

text {* Reflexive and transitive closure. *}
term "r^*"
term "r^+"
thm rtrancl_refl r_into_rtrancl rtrancl_trans

text {* An auxiliary lemma that simplifies the following proof. *}
lemma subsetI[intro!]:
  fixes A :: "('a \<times> 'b)set"
  assumes "\<And>x y. (x,y) \<in> A \<Longrightarrow> (x,y) \<in> B"
  shows "A \<subseteq> B" using assms by auto

lemma "(r O s)^-1 = s^-1 O r^-1" (is "?lhs = ?rhs")
proof
  show "?lhs \<subseteq> ?rhs"
  proof
    fix z x assume "(z,x) \<in> ?lhs"
    hence "(x,z) \<in> r O s" unfolding converse_iff .
    then obtain y where SR: "(x,y) \<in> s \<and> (y,z) \<in> r" unfolding rel_comp_def by auto
    hence "(y,z) \<in> r" by simp
    hence "(z,y) \<in> r^-1" unfolding converse_iff .
    from SR have "(x,y) \<in> s" by simp
    hence "(y,x) \<in> s^-1" unfolding converse_iff .
    with `(z,y) \<in> r^-1` have "(z,y) \<in> r^-1 \<and> (y,x) \<in> s^-1" by simp
    thus "(z,x) \<in> ?rhs" unfolding rel_comp_def by auto
  qed
next
  show "?rhs \<subseteq> ?lhs"
  proof
    fix x z assume "(x,z) \<in> ?rhs"
    then obtain y where RS: "(x,y) \<in> r^-1 \<and> (y,z) \<in> s^-1" unfolding rel_comp_def by auto
    hence "(y,z) \<in> s^-1" by simp
    hence "(z,y) \<in> s" unfolding converse_iff .
    from RS have "(x,y) \<in> r^-1" by simp
    hence "(y,x) \<in> r" unfolding converse_iff .
    with `(z,y) \<in> s` have "(z,x) \<in> r O s" unfolding rel_comp_def by auto
    thus "(x,z) \<in> ?lhs" unfolding converse_iff .
  qed
qed

section {* Inductively Defined Sets *}
inductive_set even :: "nat set"
where zero[intro!]: "0 \<in> even"
    | step[intro!]: "n \<in> even \<Longrightarrow> Suc(Suc n) \<in> even"

text {* Automatically generated lemms. *}
thm even.zero even.step even.cases even.induct

lemma two_times_even[intro!]: "2*k \<in> even" by (induct k) auto

text {* Have a look at the definition of `dvd'. *}
thm dvd_def

text {* A number that is divisible by 2 is even. *}
lemma dvd_imp_even: "2 dvd n \<Longrightarrow> n \<in> even" unfolding dvd_def by auto

text {* Have a look at the induction rule for `even'. *}
thm even.induct

text {* Every even number is divisible by 2. *}
lemma even_imp_dvd: "n \<in> even \<Longrightarrow> 2 dvd n"
proof (induct rule: even.induct)
  print_cases
  case zero show ?case by simp
next
  case (step n)
  hence IH: "2 dvd n" by simp
  then obtain k where "n = 2*k" unfolding dvd_def by (rule exE)
  hence "Suc(Suc n) = 2*(Suc k)" by simp
  thus ?case unfolding dvd_def by (rule exI[where x = "Suc k"])
qed

text {* A number is even iff it is divisible by 2. *}
lemma even_iff_dvd: "(n \<in> even) = (2 dvd n)"
proof
 assume "n \<in> even" thus "2 dvd n" by (rule even_imp_dvd)
next
 assume "2 dvd n" thus "n \<in> even" by (rule dvd_imp_even)
qed

subsection{* Advanced Inductive Sets *}
text {* The reflexive and transitive closure. *}
inductive_set rtc :: "('a \<times> 'a)set \<Rightarrow> ('a \<times> 'a)set" ("_*" [1000] 999)
  for r :: "('a \<times> 'a)set"
where rtc_refl[iff]: "(x,x) \<in> r*"
    | rtc_step: "(x,y) \<in> r \<Longrightarrow> (y,z) \<in> r* \<Longrightarrow> (x,z) \<in> r*"

text {* rtc is obviously reflexive (by rule rtc_refl). Now we show that
it is indeed transitive. *}
lemma rtc_trans: "(x,y) \<in> r* \<Longrightarrow> (y,z) \<in> r* \<Longrightarrow> (x,z) \<in> r*"
proof (induct rule: rtc.induct)
  case (rtc_refl x) thus ?case .
next
  case (rtc_step x y u)
  hence "(y,z) \<in> r*" by simp
  with `(x,y) \<in> r` show ?case by (rule rtc.rtc_step)
qed

end